#
# Copyright 2015-2016, Intel Corporation
# Copyright (c) 2016, Microsoft Corporation. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

#
# src/test/obj_redo_log/TEST6 -- unit test for redo log with valgrind
# pmemcheck tool
#
[CmdletBinding(PositionalBinding=$false)]
Param(
    [alias("d")]
    $DIR = ""
    )
$Env:UNITTEST_NAME = "obj_redo_log\TEST6"
$Env:UNITTEST_NUM = "6"


# standard unit test setup
. ..\unittest\unittest.ps1

require_test_type medium

require_build_type debug

# XXX: we don't have valgrind in Windows yet
# configure_valgrind pmemcheck force-enable
# export VALGRIND_OPTS="--mult-stores=yes"

setup

$FILE=$DIR+"\pool"
$FSIZE=1024 * 1024
$RSIZE=4

create_holey_file ($FSIZE.ToString() + "b") $FILE

expect_normal_exit $Env:EXE_DIR\obj_redo_log$Env:EXESUFFIX $FILE $RSIZE `
	C `
	s:0:0x00004200:0x11111111 `
	s:1:0x00004208:0x22222222 `
	s:2:0x00004210:0x33333333 `
	f:3:0x00004218:0x44444444 `
	P `
	r:0x00004200 `
	r:0x00004208 `
	r:0x00004210 `
	r:0x000004218 `
	s:0:0x00004300:0x55555555 `
	s:1:0x00004308:0x66666666 `
	s:2:0x00004310:0x77777777 `
	s:3:0x00004318:0x88888888 `
	F:3 `
	R `
	r:0x00004300 `
	r:0x00004308 `
	r:0x00004310 `
	r:0x00004318 `
	C

check

pass
