#
# Copyright 2015-2016, Intel Corporation
# Copyright (c) 2016, Microsoft Corporation. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

#
# src/test/obj_redo_log/TEST2 -- unit test for redo_log_check function
#
[CmdletBinding(PositionalBinding=$false)]
Param(
    [alias("d")]
    $DIR = ""
    )
$Env:UNITTEST_NAME = "obj_redo_log\TEST2"
$Env:UNITTEST_NUM = "2"


# standard unit test setup
. ..\unittest\unittest.ps1

require_test_type medium

require_build_type debug

setup

$FILE=$DIR+"\pool"
$AREASIZE=4096
$FSIZE=$AREASIZE + 32768
$RSIZE=$AREASIZE / 16

create_holey_file ($FSIZE.ToString() + "b") $FILE

expect_normal_exit $Env:EXE_DIR\obj_redo_log$Env:EXESUFFIX $FILE $RSIZE `
	C `
	s:0:0x00001010:0xabcdabcd `
	C `
	s:4:0x00001020:0xabcdabcd `
	C `
	f:0:0x00001030:0xffffaaaa `
	C `
	s:0:0x00001040:0xaaaaffff `
	f:1:0x00001050:0xbbbbeeee `
	C `
	s:1:0x00001060:0xbbbbeeee `
	C `
	f:2:0x00001070:0xccccdddd `
	f:3:0x00001080:0xddddcccc `
	C `
	f:0:0x00001090:0x99999999 `
	f:1:0x000010a0:0xaaaaaaaa `
	f:4:0x000010b0:0xbbbbbbbb `
	C `
	s:0:0x00001010:0x11111111 `
	s:1:0x00001020:0x22222222 `
	s:2:0x00001030:0x33333333 `
	s:3:0x00001040:0x44444444 `
	f:4:0x10001050:0x55555555 `
	C `
	s:0:0x00001010:0x66666666 `
	s:1:0x00001020:0x77777777 `
	s:2:0x00001030:0x88888888 `
	s:3:0x10001040:0x99999999 `
	f:4:0x00001050:0xaaaaaaaa `
	C `
	s:0:0x00000010:0x11111111 `
	s:1:0x00001020:0x22222222 `
	s:2:0x00001030:0x33333333 `
	s:3:0x00001040:0x44444444 `
	f:4:0x00001050:0x55555555 `
	C `
	s:0:0x00001010:0x66666666 `
	s:1:0x00001020:0x77777777 `
	s:2:0x00001030:0x88888888 `
	s:3:0x00001040:0x99999999 `
	f:4:0x00000050:0xaaaaaaaa `
	C

check

pass
